#include "client.hpp"
#include "common.hpp"
#include "controlService.hpp"
#include "secretManager.hpp"
#include "server.hpp"
#include "serverSideLink.hpp"
#include "service.hpp"
#include "ssl.hpp"
#include "storageHelperFactory.hpp"

#include <folly/Baton.h>
#include <folly/ConcurrentSkipList.h>
#include <folly/concurrency/ConcurrentHashMap.h>
#include <folly/executors/CPUThreadPoolExecutor.h>
#include <folly/executors/GlobalExecutor.h>
#include <folly/init/Init.h>
#include <gflags/gflags.h>
#include <glog/logging.h>
#include <google/protobuf/util/json_util.h>
#include <helpers/init.h>
#include <monitoring/monitoring.h>

#include <signal.h>

#include <iostream>

static bool validatePort(const char *flagname, std::int32_t value)
{
    if (value > 0 && value < 32768)
        return true;

    std::cerr << "Invalid value for --" << flagname << ": " << value
              << std::endl;
    return false;
}

DEFINE_int32(server_port, 5555, "data port");
DEFINE_validator(server_port, &validatePort);
DEFINE_int32(helper_workers, 100, "number of threads handling helpers' IO");
DEFINE_int32(webdav_helper_workers, 25, "number of threads handling WebDAV helpers' IO");

DEFINE_string(graphite_url, "", "url to Graphite monitoring server");
DEFINE_string(graphite_namespace_prefix, "rtransfer_link",
    "prefix for monitoring metrics");
DEFINE_uint64(
    graphite_reporting_period, 30, "reporting period for metrics (in seconds)");
DEFINE_bool(graphite_reporting_full, false, "enable full metrics reporting");

int main(int argc, char **argv)
{
    gflags::ParseCommandLineFlags(&argc, &argv, true);
    folly::init(&argc, &argv);

    auto executor = std::make_shared<folly::CPUThreadPoolExecutor>(
        std::thread::hardware_concurrency());

    folly::setCPUExecutor(executor);

    one::helpers::init();

    if (!FLAGS_graphite_url.empty()) {
        auto config = std::make_shared<
            one::monitoring::GraphiteMonitoringConfiguration>();
        config->fromGraphiteURL(FLAGS_graphite_url);
        config->namespacePrefix = FLAGS_graphite_namespace_prefix;
        config->reportingPeriod = FLAGS_graphite_reporting_period;
        config->reportingLevel = FLAGS_graphite_reporting_full
                                     ? cppmetrics::core::ReportingLevel::Full
                                     : cppmetrics::core::ReportingLevel::Basic;

        one::helpers::configureMonitoring(config, true);
    }

    rtransfer::SecretManager secretManager;
    rtransfer::ServerLinksMap serverLinks;
    rtransfer::ClientLinksMap clientLinks;
    rtransfer::StoragesMap storages;
    rtransfer::StorageHelperFactory helperFactory{
        FLAGS_helper_workers, FLAGS_webdav_helper_workers, clientLinks, serverLinks};

    // TODO: have object responsible for maintaining storagemap using
    // helperFactory
    rtransfer::link_control::ControlService controlService{
        helperFactory, clientLinks, storages, secretManager};

    rtransfer::ServiceFactory serviceFactory{storages, controlService};
    rtransfer::ServerSideLinkFactory serverSideLinkFactory{serverLinks};

    wangle::ServerSocketConfig serverSocketConfig;
    if (rtransfer::useSSL())
        serverSocketConfig.sslContextConfigs.emplace_back(
            rtransfer::sslContextConfig());

    auto server =
        rtransfer::server::makeServer(FLAGS_server_port, serverSideLinkFactory,
            serviceFactory, secretManager, serverSocketConfig);

    folly::Baton<> stopBaton;
    auto controlPipeline =
        rtransfer::link_control::newPipeline(controlService, stopBaton);
    stopBaton.wait();
}
